import { tryAtIntervals, createExponentialBackoffStrategy } from '../../../internal/async/interval-retrier.js';
import { validatePathParam } from '../../../helpers/marshalling.js';
import { API } from './api.gen.js';
import { IMAGE_TRANSIENT_STATUSES, PRIVATE_NIC_TRANSIENT_STATUSES, SERVER_TRANSIENT_STATUSES, SNAPSHOT_TRANSIENT_STATUSES, VOLUME_TRANSIENT_STATUSES } from './content.gen.js';
import { unmarshalSetImageResponse } from './marshalling.gen.js';
import { marshalSetImageRequestWithID } from './marshalling.utils.js';

const validateNotUndefined = obj => {
  if (obj === undefined) throw new TypeError(`object was found undefined`);
  return obj;
};
class InstanceV1UtilsAPI extends API {
  /**
   * Waits for {@link Image} to be in a final state.
   *
   * @param request - The request {@link GetImageRequest}
   * @param options - The waiting options
   * @returns A Promise of Image
   */
  waitForImage = (request, options) => tryAtIntervals(async () => {
    const value = await this.getImage(request).then(res => res.image);
    return {
      done: !IMAGE_TRANSIENT_STATUSES.includes(value.state),
      value
    };
  }, createExponentialBackoffStrategy(options?.minDelay ?? 1, options?.maxDelay ?? 30), options?.timeout);

  /**
   * Waits for {@link PrivateNIC} to be in a final state.
   *
   * @param request - The request {@link GetPrivateNICRequest}
   * @param options - The waiting options
   * @returns A Promise of PrivateNIC
   */
  waitForPrivateNIC = (request, options) => tryAtIntervals(async () => {
    const value = await this.getPrivateNIC(request).then(res => res.privateNic);
    return {
      done: !PRIVATE_NIC_TRANSIENT_STATUSES.includes(value.state),
      value
    };
  }, createExponentialBackoffStrategy(options?.minDelay ?? 1, options?.maxDelay ?? 30), options?.timeout);

  /**
   * Waits for {@link Server} to be in a final state.
   *
   * @param request - The request {@link GetServerRequest}
   * @param options - The waiting options
   * @returns A Promise of Server
   */
  waitForServer = (request, options) => tryAtIntervals(async () => {
    const value = await this.getServer(request).then(res => res.server);
    return {
      done: !SERVER_TRANSIENT_STATUSES.includes(value.state),
      value
    };
  }, createExponentialBackoffStrategy(options?.minDelay ?? 1, options?.maxDelay ?? 30), options?.timeout);

  /**
   * Waits for {@link Server} to be in a final state.
   *
   * @param request - The request {@link GetServerRequest}
   * @param options - The waiting options
   * @returns A Promise of Server
   */
  waitForSnapshot = (request, options) => tryAtIntervals(async () => {
    const value = await this.getSnapshot(request).then(res => res.snapshot);
    return {
      done: !SNAPSHOT_TRANSIENT_STATUSES.includes(value.state),
      value
    };
  }, createExponentialBackoffStrategy(options?.minDelay ?? 1, options?.maxDelay ?? 30), options?.timeout);

  /**
   * Waits for {@link Volume} to be in a final state.
   *
   * @param request - The request {@link GetVolumeRequest}
   * @param options - The waiting options
   * @returns A Promise of Volume
   */
  waitForVolume = (request, options) => tryAtIntervals(async () => {
    const value = await this.getVolume(request).then(res => res.volume);
    return {
      done: !VOLUME_TRANSIENT_STATUSES.includes(value.state),
      value
    };
  }, createExponentialBackoffStrategy(options?.minDelay ?? 1, options?.maxDelay ?? 30), options?.timeout);

  /**
   * Updates a snapshot.
   *
   * @param request - The request {@link UpdateSnapshotRequest}
   * @returns A Promise of UpdateSnapshotResponse
   */
  updateSnapshot = request => this.getSnapshot(request).then(res => validateNotUndefined(res.snapshot)).then(snapshot => this._setSnapshot({
    ...snapshot,
    name: request.name ?? snapshot.name,
    snapshotId: snapshot.id
  })).then(res => ({
    snapshot: res.snapshot
  }));

  /**
   * Updates a security group.
   *
   * @param request - The request {@link UpdateSecurityGroupRequest}
   * @returns A Promise of UpdateSecurityGroupResponse
   */
  updateSecurityGroup = request => this.getSecurityGroup({
    securityGroupId: request.securityGroupId,
    zone: request.zone
  }).then(res => validateNotUndefined(res.securityGroup)).then(securityGroup => this._setSecurityGroup({
    creationDate: securityGroup.creationDate,
    description: request.description ?? securityGroup.description,
    enableDefaultSecurity: request.enableDefaultSecurity ?? securityGroup.enableDefaultSecurity,
    id: securityGroup.id,
    inboundDefaultPolicy: request.inboundDefaultPolicy ?? securityGroup.inboundDefaultPolicy,
    modificationDate: securityGroup.modificationDate,
    name: request.name ?? securityGroup.name,
    organization: securityGroup.organization,
    organizationDefault: request.organizationDefault ?? securityGroup.organizationDefault,
    outboundDefaultPolicy: request.outboundDefaultPolicy ?? securityGroup.outboundDefaultPolicy,
    project: securityGroup.project,
    projectDefault: request.projectDefault ?? securityGroup.projectDefault,
    servers: securityGroup.servers,
    stateful: request.stateful ?? securityGroup.stateful,
    zone: request.zone
  })).then(res => ({
    securityGroup: res.securityGroup
  }));

  /**
   * Updates a security group rule.
   *
   * @param request - The request {@link UpdateSecurityGroupRuleRequest}
   * @returns A Promise of UpdateSecurityGroupRuleResponse
   */
  updateSecurityGroupRule = request => this.getSecurityGroupRule({
    securityGroupId: request.securityGroupId,
    securityGroupRuleId: request.securityGroupRuleId,
    zone: request.zone
  }).then(res => validateNotUndefined(res.rule)).then(rule => {
    let sReq = {
      action: request.action ?? rule.action,
      destPortFrom: rule.destPortFrom,
      destPortTo: rule.destPortTo,
      direction: request.direction ?? rule.direction,
      editable: rule.editable,
      id: request.securityGroupRuleId,
      ipRange: request.ipRange ?? rule.ipRange,
      position: request.position ?? rule.position,
      protocol: request.protocol ?? rule.protocol,
      securityGroupId: request.securityGroupId,
      securityGroupRuleId: request.securityGroupRuleId
    };
    if (request.destPortFrom) {
      sReq = {
        ...sReq,
        destPortFrom: request.destPortFrom > 0 ? request.destPortFrom : undefined
      };
    }
    if (request.destPortTo) {
      sReq = {
        ...sReq,
        destPortTo: request.destPortTo > 0 ? request.destPortTo : undefined
      };
    }
    if (sReq.destPortFrom && sReq.destPortTo && sReq.destPortFrom === sReq.destPortTo) {
      sReq = {
        ...sReq,
        destPortTo: undefined
      };
    }
    // When we use ICMP protocol portFrom and portTo should be set to nil
    if (request.protocol === 'ICMP') {
      sReq = {
        ...sReq,
        destPortFrom: undefined,
        destPortTo: undefined
      };
    }
    return this._setSecurityGroupRule(sReq);
  }).then(res => ({
    rule: res.rule
  }));

  /**
   * Updates a server.
   *
   * @param request - The request {@link UpdateServerRequest}
   * @returns A Promise of UpdateServerResponse
   */
  updateServer = request => this._updateServer(request);

  /**
   * Creates a server.
   *
   * @param request - The request {@link CreateServerRequest}
   * @returns A Promise of CreateServerResponse
   */
  createServer = request => this._createServer(request);

  /**
   * Starts an action and wait for the server to be in the correct "terminal
   * state" expected by this action.
   *
   * @param request - The request {@link ServerActionRequest}
   * @returns A Promise of Server
   */
  serverActionAndWait = async (request, options) => {
    const finalServer = await this.serverAction(request).then(() => this.waitForServer({
      serverId: request.serverId,
      zone: request.zone
    }, options));

    // Check the action was properly executed.
    let expectedState;
    switch (request.action) {
      case 'poweron':
      case 'reboot':
        expectedState = 'running';
        break;
      case 'poweroff':
        expectedState = 'stopped';
        break;
      case 'stop_in_place':
        expectedState = 'stopped in place';
        break;
    }
    if (expectedState && finalServer.state !== expectedState) {
      throw new Error(`expected state ${expectedState} but found ${finalServer.state}: ${finalServer.stateDetail}`);
    }
    return finalServer;
  };

  /**
   * Attaches a volume to a server.
   *
   * @param request - The request {@link AttachVolumeRequest}
   * @returns A Promise of AttachVolumeResponse
   */
  attachVolume = async request => {
    const volumes = await this.getServer({
      serverId: request.serverId,
      zone: request.zone
    }).then(res => validateNotUndefined(res.server?.volumes));
    const newVolumes = {};
    for (const [key, server] of Object.entries(volumes)) {
      newVolumes[key] = {
        id: server.id,
        name: server.name
      };
    }

    // We loop through all the possible volume keys (0 to len(volumes))
    // to find a non existing key and assign it to the requested volume.
    // A key should always be found. However we return an error if no keys were found.
    let found = false;
    const volumesLength = Object.keys(volumes).length;
    for (let index = 0; index <= volumesLength; index += 1) {
      const key = index.toString();
      if (!(key in newVolumes)) {
        newVolumes[key] = {
          id: request.volumeId,
          name: request.volumeId // name is ignored on this PATCH
        };

        found = true;
        break;
      }
    }
    if (!found) {
      throw new Error(`could not find key to attach volume ${request.volumeId}`);
    }

    // Update server
    return this.updateServer({
      serverId: request.serverId,
      volumes: newVolumes,
      zone: request.zone
    }).then(obj => obj);
  };

  /**
   * Detaches a volume from a server.
   *
   * @param request - The request {@link DetachVolumeRequest}
   * @returns A Promise of DetachVolumeResponse
   */
  detachVolume = async request => {
    // Get server and related volumes.
    const server = await this.getVolume({
      volumeId: request.volumeId,
      zone: request.zone
    }).then(res => validateNotUndefined(res.volume?.server?.id)).then(serverId => this.getServer({
      serverId,
      zone: request.zone
    })).then(res => validateNotUndefined(res.server));

    // Remove volume.
    const newVolumes = {};
    for (const [key, volume] of Object.entries(server.volumes)) {
      if (volume.id !== request.volumeId) {
        newVolumes[key] = {
          id: volume.id,
          name: volume.name
        };
      }
    }

    // Update server.
    return this.updateServer({
      serverId: server.id,
      volumes: newVolumes,
      zone: request.zone
    }).then(obj => obj);
  };

  /**
   * Updates an image.
   *
   * @param request - The request {@link UpdateImageRequest}
   * @returns A Promise of UpdateImageResponse
   */
  updateImage = request => this.getImage({
    zone: request.zone,
    imageId: request.imageId
  }).then(res => validateNotUndefined(res.image)).then(image => ({
    ...image,
    name: request.name ?? image.name,
    tags: request.tags ?? image.tags,
    id: image.id
  })).then(imageReq => this.client.fetch({
    body: JSON.stringify(marshalSetImageRequestWithID(imageReq, this.client.settings)),
    headers: {
      'Content-Type': 'application/json; charset=utf-8'
    },
    method: 'PUT',
    path: `/instance/v1/zones/${validatePathParam('zone', imageReq.zone)}/images/${validatePathParam('id', imageReq.id)}`
  }, unmarshalSetImageResponse)).then(res => ({
    image: res.image
  }));

  /**
   * Get the content of a user data on a server for the given key.
   *
   * @param request - The request {@link GetServerUserDataRequest}
   * @returns The content of the key
   */
  getServerUserData = request => this.client.fetch({
    method: 'GET',
    path: `/instance/v1/zones/${validatePathParam('zone', request.zone ?? this.client.settings.defaultZone)}/servers/${validatePathParam('serverId', request.serverId)}/user_data/${validatePathParam('key', request.key)}`
  });

  /**
   * Sets the content of a user data on a server for the given key.
   *
   * @param request - The request {@link SetServerUserDataRequest}
   */
  setServerUserData = request => this.client.fetch({
    body: request.content,
    headers: {
      'Content-Type': 'text/plain'
    },
    method: 'PATCH',
    path: `/instance/v1/zones/${validatePathParam('zone', request.zone ?? this.client.settings.defaultZone)}/servers/${validatePathParam('serverId', request.serverId)}/user_data/${validatePathParam('key', request.key)}`
  });
}

export { InstanceV1UtilsAPI };
