import { isJSONObject } from '../../../helpers/json.js';
import { ScalewayError } from '../scw-error.js';

/**
 * Details of an {@link PermissionsDeniedError} error.
 *
 * @public
 */

/**
 * Build the default message for {@link PermissionsDeniedError}.
 *
 * @param list - The list of {@link PermissionsDeniedErrorDetails}
 * @returns The error message
 *
 * @internal
 */
const buildMessage = list => `insufficient permissions: ${list.map(_ref => {
  let {
    action,
    resource
  } = _ref;
  return `${action} ${resource}`;
}).join('; ')}`;

/**
 * PermissionsDenied error happens when one or many permissions are not accorded to the user making the request.
 *
 * @public
 */
class PermissionsDeniedError extends ScalewayError {
  constructor(status, body, list) {
    super(status, body, buildMessage(list));
    this.status = status;
    this.body = body;
    this.list = list;
    this.name = 'PermissionsDeniedError';
  }
  static fromJSON(status, obj) {
    if (!Array.isArray(obj.details)) return null;
    return new PermissionsDeniedError(status, obj, obj.details.reduce((list, detail) => isJSONObject(detail) && typeof detail.resource === 'string' && typeof detail.action === 'string' ? list.concat({
      action: detail.action,
      resource: detail.resource
    }) : list, []));
  }
}

export { PermissionsDeniedError };
