import { isJSONObject } from '../../../helpers/json.js';
import { ScalewayError } from '../scw-error.js';

/**
 * Scope of an {@link QuotasExceededErrorDetails} error.
 *
 * @public
 */

/**
 * Details of an {@link QuotasExceededError} error.
 *
 * @public
 */

/**
 * Build the default message for {@link QuotasExceededError}.
 *
 * @param list - The list of {@link QuotasExceededErrorDetails}
 * @returns The error message
 *
 * @internal
 */
const buildMessage = list => `quota(s) exceeded: ${list.map(details => {
  const message = `${details.resource} has reached its quota (${details.current}/${details.quota})`;
  return details.scope ? `${message} for ${details.scope.kind} '${details.scope.id}'` : message;
}).join('; ')}`;
const buildScope = detail => {
  if (typeof detail.organization_id === 'string' && detail.organization_id.length) {
    return {
      id: detail.organization_id,
      kind: 'organization'
    };
  }
  if (typeof detail.project_id === 'string' && detail.project_id.length) {
    return {
      id: detail.project_id,
      kind: 'project'
    };
  }
  return undefined;
};

/**
 * QuotasExceeded error happens when one or many resource exceed quotas during the creation of a resource.
 *
 * @public
 */
class QuotasExceededError extends ScalewayError {
  constructor(status, body, list) {
    super(status, body, buildMessage(list));
    this.status = status;
    this.body = body;
    this.list = list;
    this.name = 'QuotasExceededError';
  }
  static fromJSON(status, obj) {
    if (!Array.isArray(obj.details)) return null;
    return new QuotasExceededError(status, obj, obj.details.reduce((list, detail) => isJSONObject(detail) && typeof detail.resource === 'string' && typeof detail.quota === 'number' && typeof detail.current === 'number' ? list.concat({
      current: detail.current,
      quota: detail.quota,
      resource: detail.resource,
      scope: buildScope(detail)
    }) : list, []));
  }
}

export { QuotasExceededError };
