const extract = key => result => result[key];
function* pages(key, fetcher, request, firstPage) {
  if (!Array.isArray(firstPage[key])) {
    throw new Error(`Property ${key} is not a list in paginated result`);
  }
  const getList = extract(key);
  let page = request.page || 1;
  if (page === 1) {
    yield Promise.resolve(getList(firstPage));
    page += 1;
  }
  const {
    length
  } = firstPage[key];
  if (!length) return;
  const {
    totalCount
  } = firstPage;
  while (page <= Math.floor((totalCount + length - 1) / length)) {
    yield fetcher({
      ...request,
      page
    }).then(getList);
    page += 1;
  }
}

/**
 * Fetches a paginated resource.
 *
 * @param key - The resource key of values list
 * @param fetcher - The method to retrieve paginated resources
 * @param request - A request with pagination options
 * @param initial - The first page
 * @returns An async generator of resources arrays
 */
function fetchPaginated(key, fetcher, request, initial) {
  try {
    if (initial === void 0) {
      initial = fetcher(request);
    }
    return async function* () {
      yield* pages(key, fetcher, request, await initial);
    }();
  } catch (e) {
    return Promise.reject(e);
  }
}

/**
 * Fetches all paginated resource.
 *
 * @param key - The resource key of values list
 * @param fetcher - The method to retrieve paginated resources
 * @param request - A request with pagination options
 * @param initial - The first page
 * @returns A resources array Promise
 */
const fetchAll = async function (key, fetcher, request, initial) {
  if (initial === void 0) {
    initial = fetcher(request);
  }
  return (await Promise.all(Array.from(pages(key, fetcher, request, await initial)))).flat();
};

/**
 * Enriches a listing method with helpers.
 *
 * @param key - The resource key of values list
 * @param fetcher - The method to retrieve paginated resources
 * @param request - A request with pagination options
 * @returns A resource Promise with the pagination helpers
 *
 * @internal
 */
const enrichForPagination = (key, fetcher, request) => {
  const firstPage = fetcher(request);
  return Object.assign(firstPage, {
    all: () => fetchAll(key, fetcher, request, firstPage),
    [Symbol.asyncIterator]: () => fetchPaginated(key, fetcher, request, firstPage)
  });
};

export { enrichForPagination, extract, fetchAll, fetchPaginated };
